/*
 * location.c - Source code location of language elements
 *
 * Written 2001 by Werner Almesberger
 * Copyright 2001 EPFL-ICA, Network Robots
 */

/*
 * KNOWN BUG: outputs tunnels once per selector @@@
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "util.h"
#include "error.h"
#include "data.h"
#include "tree.h"
#include "device.h"
#include "police.h"
#include "filter.h"
#include "qdisc.h"
#include "location.h"


char *file_name = NULL;
int lineno = 1;

static int file_name_ref = 0; /* reference count for "file_name" */


/* ----- Registry ---------------------------------------------------------- */


void input_file(const char *name)
{
    if (file_name && !file_name_ref) free(file_name);
    file_name = stralloc(name);
    file_name_ref = 0;
}


LOCATION current_location(void)
{
    LOCATION loc;

    loc.file = file_name ? file_name : "(stdin)";
    loc.tag = NULL;
    loc.line = lineno;
    file_name_ref++;
    return loc;
}


int print_location(FILE *file,LOCATION loc)
{
    return fprintf(file,"%s:%d",loc.file,loc.line);
}


int print_current_location(FILE *file)
{
    return fprintf(file,"%s:%d",file_name ? file_name : "(stdin)",lineno);
}


/* ----- File output helper functions -------------------------------------- */


static FILE *file_open(const char *name)
{
    FILE *file;

    fflush(stdout);
    fflush(stderr);
    if (!strcmp(name,"stderr")) return stderr;
    file = fopen(name,"w");
    if (file) return file;
    perror(file_name);
    exit(1);
}


static void file_close(FILE *file,const char *name)
{
    fflush(file);
    if (ferror(file)) errorf("error writing to \"%s\"",name);
    if (strcmp(name,"stderr"))
	if (fclose(file) == EOF) {
	    perror(name);
	    exit(1);
	}
}


/* ----- Dump location map ------------------------------------------------- */


static void do_map(FILE *file,LOCATION loc)
{
    const char *file_name,*walk;

    file_name = loc.file && *loc.file ? loc.file : "-";
    for (walk = file_name; *walk; walk++)
	if (*walk <= ' ' || *walk > '~') {
	    file_name = "<unprintable>";
	    break;
	}
    fprintf(file," %s %s %d\n",loc.tag ? loc.tag : "-",file_name,loc.line);
}


static void id_qdisc(FILE *file,const QDISC *qdisc)
{
    fprintf(file,"%s:",qdisc->parent.device->name);
    if (qdisc->dsc == &ingress_dsc) fprintf(file,"ingress");
    else fprintf(file,"%u",(unsigned) qdisc->number);
}


static void id_class(FILE *file,const CLASS *class)
{
     id_qdisc(file,class->parent.qdisc);
     fprintf(file,":%u",(unsigned) class->number);
}


static void id_filter(FILE *file,const FILTER *filter)
{
    if (filter->parent.class) id_class(file,filter->parent.class);
    else {
	id_qdisc(file,filter->parent.qdisc);
	fprintf(file,":");
    }
    fprintf(file,":%u",(unsigned) filter->number);
}


static void map_elements(FILE *file,const ELEMENT *elements)
{
    const ELEMENT *element;
    int n = 0;

    for (element = elements; element; element = element->next) {
	fprintf(file,"element ");
	id_filter(file,element->parent.filter);
	fprintf(file,":%d",n);
	do_map(file,element->location);
	if (element->parent.class == &class_is_tunnel) {
	    const ELEMENT *scan;

	    for (scan = elements; scan != element; scan = scan->next)
		if (scan->parent.class == &class_is_tunnel &&
		  scan->parent.tunnel == element->parent.tunnel)
		    break;
	    if (scan == element) {
		fprintf(file,"tunnel ");
		id_filter(file,element->parent.filter);
		fprintf(file,":%u",(unsigned) element->parent.tunnel->number);
		do_map(file,element->parent.tunnel->location);
	    }
	}
	n++;
    }
}


static void map_filters(FILE *file,const FILTER *filters)
{
    const FILTER *filter;

    for (filter = filters; filter; filter = filter->next) {
	fprintf(file,"filter ");
	id_filter(file,filter);
	do_map(file,filter->location);
	map_elements(file,filter->elements);
    }
}


static void map_qdisc(FILE *file,const QDISC *qdisc);


static void map_classes(FILE *file,const CLASS *classes)
{
    const CLASS *class;

    for (class = classes; class; class = class->next) {
	fprintf(file,"class ");
	id_class(file,class);
	do_map(file,class->location);
	map_classes(file,class->child);
	map_qdisc(file,class->qdisc);
	map_filters(file,class->filters);
    }
}


static void map_qdisc(FILE *file,const QDISC *qdisc)
{
    if (!qdisc) return;
    fprintf(file,"qdisc ");
    id_qdisc(file,qdisc);
    do_map(file,qdisc->location);
    map_classes(file,qdisc->classes);
    map_filters(file,qdisc->filters);
}


static void map_devices(FILE *file)
{
    const DEVICE *device;

    for (device = devices; device; device = device->next) {
	fprintf(file,"device %s",device->name);
	do_map(file,device->location);
	map_qdisc(file,device->ingress);
	map_qdisc(file,device->egress);
    }
}


static void map_policers(FILE *file)
{
    const POLICE *p;

    for (p = policers; p; p = p->next)
	if (p->used) {
	    fprintf(file,"police %u",(unsigned) p->number);
	    do_map(file,p->location);
	}
}


void write_location_map(const char *file_name)
{
    FILE *file;

    file = file_open(file_name);
    map_devices(file);
    map_policers(file);
    file_close(file,file_name);
}


/* ----- Dump variable use information ------------------------------------- */


typedef struct _var_register {
    const char *name;
    DATA data;
    struct _var_register *next;
} VAR_REGISTER;

static VAR_REGISTER *var_register = NULL,**last_var = &var_register;


void register_var(const char *name,DATA d)
{
    VAR_REGISTER *new;

    new = alloc_t(VAR_REGISTER);
    new->name = stralloc(name);
    new->data = d;
    new->next = NULL;
    *last_var = new;
    last_var = &new->next;
}


static void dump_var_use(FILE *file,const VAR_REGISTER *var)
{
    fprintf(file,"$%s = ",var->name);
    switch (var->data.type) {
	case dt_qdisc:
	    fprintf(file,"qdisc ");
	    id_qdisc(file,var->data.u.qdisc);
	    break;
	case dt_class:
	    fprintf(file,"class ");
	    id_class(file,var->data.u.class);
	    break;
	case dt_filter:
	    fprintf(file,"filter");
	    id_filter(file,var->data.u.filter);
	    break;
	case dt_police:
	    fprintf(file,"police %u",(unsigned) var->data.u.police->number);
	    break;
	default:
	    print_data(file,var->data);
    }
    fputc('\n',file);
}


void write_var_use(const char *file_name)
{
    FILE *file;

    file = file_open(file_name);
    while (var_register) {
	VAR_REGISTER *next;

	dump_var_use(file,var_register);
	next = var_register->next;
	free((char *) var_register->name);
	free(var_register);
	var_register = next;
    }
    file_close(file,file_name);
}
